# Authors: Hari Raval, Zongjun Liu, Aditi Memani, Joseph Zuccarelli             #
# Course: AC 207                                                                #
# File: optimzers.py                                                            #
# Description: Create six different optimization techniques that leverage       #
# forward mode automatic differentiation, enabling a user to find the minimum   #
# value of a function, location of the minimum value, and wall clock time to    #
# find these values.                                                            #
#################################################################################

import numpy as np
from ad_AHJZ.forward_mode import forward_mode
import time


class Optimizer:
    """
    A class containing six different optimizer methods that leverage forward mode
    automatic differentiation, enabling a user to to find the minimum value of a function,
    location of the minimum value, and wall clock time to find these values

    Examples
    --------
    # sample use case of instantiating and using an adam optimizer
    >>> x = 1
    >>> opt = Optimizer()
    >>> f_x = lambda x: x** 2
    >>> print(opt.ADAM(x, f_x, 1000))
    (0.058393001556396484, 3.2900616291354705e-42, array([-1.8138527e-21]))

    # sample use case of instantiating and using a momentum optimizer
    >>> x = 2
    >>> opt = Optimizer()
    >>> f_x = lambda x: x** 4
    >>> print(opt.momentum(x, f_x, 1000))
    (0.04654884338378906, 2.0278841795288425e-05, array([-0.06710591]))

    """

    @staticmethod
    def momentum(x, f_x, num_iter, alpha=0.01, beta=.9):
        """
        Parameters
        ----------
        x: the variable input (can be in either scalar or vector form)
        f_x: the function you would like to obtain the minimum for
        num_iter: the number of interations to perform
        alpha: learning rate for the gradiant descent (default 0.01)
        beta: exponential decay (default 0.9)

        Returns
        -------
        opt_time: The time it takes to run the optimizer in seconds
        val: the position of the minimum value
        curr_val: the minimum value (can be in either scalar or vector form)

        Examples
        --------
        >>> x = 1
        >>> f_x = lambda x: (-1 * x.log()) + (x.exp() * x**4) / 10
        >>> Optimizer.momentum(x, f_x, 1000)
        (0.24898195266723633, 0.26172998379097046, array([0.94233316]))

        >>> x = np.array([1, -1])
        >>> f_x = lambda x, y:x**3 + y**2
        >>> Optimizer.momentum(x, f_x, 1000)
        (0.08585095405578613, 2.7605629377339922e-05, array([ 3.02226506e-02, -3.30135704e-12]))

        >>> x = 2
        >>> f_x = lambda x: (x - 1)**2 + 5
        >>> Optimizer.momentum(x, f_x, 1000)
        (0.06420779228210449, 5.0, array([1.]))

        """

        # start the timer
        start = time.time()
        # decay value must be between 0 and 1
        if 0 <= beta < 1:
            mt, curr_val = 0, x
            fm = forward_mode(x, f_x)
            val, x_der = fm.get_function_value(), fm.get_jacobian()

            # perform momentum optimization for the number of iterations specified
            for t in range(1, num_iter + 1):
                mt = beta * mt + (1 - beta) * x_der
                variation = alpha * mt
                # compute the new variation to update the current x location
                curr_val = curr_val - variation
                # recalculate the function value and derivative at the updated value
                fm = forward_mode(curr_val, f_x)
                val, x_der = fm.get_function_value(), fm.get_jacobian()
        # raise the appropriate error for beta value not within 0 to 1
        else:
            raise ValueError("Beta Values must be within the range of [0,1)")

        # end the timer and compute wall clock time
        end = time.time()
        opt_time = end - start
        return opt_time, val, curr_val

    @staticmethod
    def ADAM(x, f_x, num_iter, alpha=0.01, beta1=.9, beta2=.999, epsilon=1e-8):
        """
        Parameters
        ----------
        x: the variable input (can be in either scalar or vector form)
        f_x: the function you would like to obtain the minimum for
        num_iter: the number of interations to perform
        alpha: learning rate for the gradiant descent (default 0.01)
        beta1: exponential decay of the first moment (default 0.90)
        beta2: exponential decay of the second moment (default 0.999)
        epsilon: denominator value to assure that ZeroDivisionError is not raised (default 1e-8)

        Returns
        -------
        opt_time: The time it takes to run the optimizer in seconds
        val: the position of the minimum value
        curr_val: the minimum value (can be in either scalar or vector form)

        Examples
        --------
        >>> x = 1
        >>> f_x = lambda x: (-1 * x.log()) + (x.exp() * x**4) / 10
        >>> Optimizer.ADAM(x, f_x, 1000)
        (0.19765210151672363, 0.26172998379097046, array([0.94233316]))

        >>> x = np.array([1,-1])
        >>> f_x = lambda x,y:x**3 + y**2
        >>> Optimizer.ADAM(x, f_x, 1000)
        (0.09849190711975098, 6.03886825409073e-06, array([1.82103595e-02, 1.81385270e-21]))

        >>> x = 2
        >>> f_x = lambda x: (x-1)**2 + 5
        >>> Optimizer.ADAM(x, f_x, 1000)
        (0.08600378036499023, 5.0, array([1.]))

        """

        # start the timer
        start = time.time()

        # decay values must be between 0 and 1
        if 0 <= beta1 < 1:
            if 0 <= beta2 < 1:
                # initialize moment and direction variables
                mt, vt = 0, 0
                curr_val = x
                fm = forward_mode(x, f_x)
                val, x_der = fm.get_function_value(), fm.get_jacobian()

                # perform ADAM optimization for the number of iterations specified
                for t in range(1, num_iter + 1):
                    mt = beta1 * mt + (1 - beta1) * x_der
                    vt = beta2 * vt + (1 - beta2) * x_der ** 2
                    mhat = mt / (1 - beta1 ** t)
                    vhat = vt / (1 - beta2 ** t)
                    # compute the new variation to update the current x location
                    variation = alpha * mhat / (np.sqrt(vhat) + epsilon)
                    curr_val = curr_val - variation
                    # recalculate the function value and derivative at the updated value
                    fm = forward_mode(curr_val, f_x)
                    val, x_der = fm.get_function_value(), fm.get_jacobian()
            # raise the appropriate error for beta 1 value not within 0 to 1
            else:
                raise ValueError("Beta Values must be within the range of [0,1)")
        # raise the appropriate error for beta 2 value not within 0 to 1
        else:
            raise ValueError("Beta Values must be within the range of [0,1)")

        # end the timer and compute wall clock time
        end = time.time()
        opt_time = end - start
        return opt_time, val, curr_val

    @staticmethod
    def NADAM(x, f_x, num_iter, alpha=0.01, beta1=.9, beta2=.999, epsilon=1e-8):
        """
        Parameters
        ----------
        x: the variable input (can be in either scalar or vector form)
        f_x: the function you would like to obtain the minimum for
        num_iter: the number of interations to perform
        alpha: learning rate for the gradiant descent (default 0.01)
        beta1: exponential decay of the first moment (default 0.90)
        beta2: exponential decay of the second moment (default 0.999)
        epsilon: denominator value to assure that ZeroDivisionError is not raised (default 1e-8)

        Returns
        -------
        opt_time: The time it takes to run the optimizer in seconds
        val: the position of the minimum value
        curr_val: the minimum value (can be in either scalar or vector form)

        Examples
        --------
        >>> x = 1
        >>> f_x = lambda x: (-1 * x.log()) + (x.exp() * x**4) / 10
        >>> Optimizer.NADAM(x, f_x, 1000)
        (0.17513608932495117, 0.26172998379097046, array([0.94233316]))

        >>> x = np.array([1,-1])
        >>> f_x = lambda x, y:x**3 + y**2
        >>> Optimizer.NADAM(x, f_x, 1000)
        (0.09782099723815918, 6.246798742956698e-06, array([ 1.8417012e-02, -2.7649170e-21]))

        >>> x = 2
        >>> f_x = lambda x: (x-1)**2 + 5
        >>> Optimizer.NADAM(x, f_x, 1000)
        (0.07419586181640625, 5.0, array([1.]))

        """

        # start the timer
        start = time.time()
        # decay values must be between 0 and 1
        if 0 <= beta1 < 1:
            if 0 <= beta2 < 1:
                # initialize moment and direction variables
                mt, vt = 0, 0
                curr_val = x
                fm = forward_mode(x, f_x)
                val, x_der = fm.get_function_value(), fm.get_jacobian()

                # perform NADAM optimization for the number of iterations specified
                for t in range(1, num_iter + 1):
                    mt = beta1 * mt + (1 - beta1) * x_der
                    vt = beta2 * vt + (1 - beta2) * x_der ** 2
                    mhat = mt / (1 - beta1 ** t)
                    vhat = vt / (1 - beta2 ** t)
                    mvar = beta1 * mhat + (1 - beta1) * x_der
                    # compute the new variation to update the current x location (note: epsilon is under square root)
                    variation = alpha * mvar / (np.sqrt(vhat + epsilon))
                    curr_val = curr_val - variation
                    # recalculate the function value and derivative at the updated value
                    fm = forward_mode(curr_val, f_x)
                    val, x_der = fm.get_function_value(), fm.get_jacobian()
            # raise the appropriate error for beta 1 value not within 0 to 1
            else:
                raise ValueError("Beta Values must be within the range of [0,1)")
        # raise the appropriate error for beta 2 value not within 0 to 1
        else:
            raise ValueError("Beta Values must be within the range of [0,1)")

        # end the timer and compute wall clock time
        end = time.time()
        opt_time = end - start
        return opt_time, val, curr_val

    @staticmethod
    def RMSprop(x, f_x, num_iter, alpha=0.01, beta=.9, epsilon=1e-8):
        """
        Parameters
        ----------
        x: the variable input (can be in either scalar or vector form)
        f_x: the function you would like to obtain the minimum for
        num_iter: the number of interations to perform
        alpha: learning rate for the gradiant descent (default 0.01)
        beta: exponential decay (default 0.9)
        epsilon: denominator value to assure that ZeroDivisionError is not raised (default 1e-8)

        Returns
        -------
        opt_time: The time it takes to run the optimizer in seconds
        val: the position of the minimum value
        curr_val: the minimum value (can be in either scalar or vector form)

        Examples
        --------
        >>> x = 1
        >>> f_x = lambda x: (-1 * x.log()) + (x.exp() * x**4) / 10
        >>> Optimizer.RMSprop(x, f_x, 1000)
        (0.3034090995788574, 0.2618028370373199, array([0.93730206]))

        >>> x = np.array([1,-1])
        >>> f_x = lambda x, y:x**3 + y**2
        >>> Optimizer.RMSprop(x, f_x, 1000)
        (0.08888602256774902, 2.4997500000081948e-05, array([ 4.34424616e-06, -4.99974999e-03]))

        >>> x = 2
        >>> f_x = lambda x: (x - 1)**2 + 5
        >>> Optimizer.RMSprop(x, f_x, 1000)
        (0.07326507568359375, 5.0, array([1.]))

        """

        # start the timer
        start = time.time()
        # decay value must be between 0 and 1
        if 0 <= beta < 1:
            vt, curr_val = 0, x
            fm = forward_mode(x, f_x)
            val, x_der = fm.get_function_value(), fm.get_jacobian()

            # perform RMSprop optimization for the number of iterations specified
            for t in range(1, num_iter + 1):
                vt = beta * vt + (1 - beta) * x_der ** 2
                variation = alpha * x_der / (np.sqrt(vt + epsilon))
                # compute the new variation to update the current x location
                curr_val = curr_val - variation
                # recalculate the function value and derivative at the updated value
                fm = forward_mode(curr_val, f_x)
                val, x_der = fm.get_function_value(), fm.get_jacobian()
        # raise the appropriate error for beta value not within 0 to 1
        else:
            raise ValueError("Beta Values must be within the range of [0,1)")

        # end the timer and compute wall clock time
        end = time.time()
        opt_time = end - start
        return opt_time, val, curr_val

    @staticmethod
    def broyden(x, f_x, num_iter, alpha=0.01):
        """
        Parameters
        ----------
        x: the variable input (can be in either scalar or vector form)
        f_x: the function you would like to obtain the minimum for
        num_iter: the number of interations to perform
        alpha: learning rate for the gradiant descent (default 0.01)

        Returns
        -------
        opt_time: The time it takes to run the optimizer in seconds
        val: the position of the minimum value
        curr_val: the minimum value (can be in either scalar or vector form)

        Examples
        --------
        >>> x = 1
        >>> f_x = lambda x: (-1 * x.log()) + (x.exp() * x**4) / 10
        >>> Optimizer.broyden(x, f_x, 1000)
        (0.2750861644744873, 0.2617299838095016, array([0.94233569]))

        >>> x = np.array([1, -1])
        >>> f_x = lambda x, y:x**3 + y**2
        >>> Optimizer.broyden(x, f_x, 1000)
        (0.1135101318359375, 3.060640040173604e-07, array([ 6.72566138e-03, -4.28010827e-05]))

        >>> x = 2
        >>> f_x = lambda x: (x - 1)**2 + 5
        >>> Optimizer.broyden(x, f_x, 1000)
        broyden: (0.0851430892944336, 5.000000001826295, array([1.00004274]))

        """

        # start the timer
        start = time.time()
        if np.isscalar(x):
            n = 1
        else:
            n = len(x)
        curr_val = x
        # start with the hessian as the identity matrix
        curr_hessian = np.eye(n)
        fm = forward_mode(x, f_x)
        val, x_der = fm.get_function_value(), fm.get_jacobian()

        # perform broyden optimization for the number of iterations specified
        for t in range(1, num_iter + 1):
            # compute the new variation to update the current x location
            variation = -alpha * curr_hessian @ x_der
            curr_val = curr_val + variation
            # recalculate the function value and derivative at the updated value
            fm = forward_mode(curr_val, f_x)
            val, x_der1 = fm.get_function_value(), fm.get_jacobian()
            yk = (x_der1 - x_der).reshape(-1, 1)
            xk = variation.reshape(-1, 1)
            # compute the amount the hessian will change by
            broyden_numerator = (xk - curr_hessian @ yk) @ xk.T @ curr_hessian
            broyden_denom = xk.T @ curr_hessian @ yk
            # update the hessian matrix per iteration (second-order method)
            curr_hessian = curr_hessian + broyden_numerator / broyden_denom
            x_der = x_der1

        # end the timer and compute wall clock time
        end = time.time()
        opt_time = end - start
        return opt_time, val, curr_val

    @staticmethod
    def BFGS(x, f_x, num_iter, alpha=0.01):
        """
        Parameters
        ----------
        x: the variable input (can be in either scalar or vector form)
        f_x: the function you would like to obtain the minimum for
        num_iter: the number of interations to perform
        alpha: learning rate for the gradiant descent (default 0.01)

        Returns
        -------
        opt_time: The time it takes to run the optimizer in seconds
        val: the position of the minimum value
        curr_val: the minimum value (can be in either scalar or vector form)

        Examples
        --------
        >>> x = 1
        >>> f_x = lambda x: (-1 * x.log()) + (x.exp() * x**4) / 10
        >>> Optimizer.BFGS(x, f_x, 1000)
        (0.2875189781188965, 0.2617299838095016, array([0.94233569]))

        >>> x = np.array([1, -1])
        >>> f_x = lambda x, y:x**3 + y**2
        >>> Optimizer.BFGS(x, f_x, 1000)
        (0.12035322189331055, 2.864368146437936e-07, array([ 6.57872856e-03, -4.13716985e-05]))

        >>> x = 2
        >>> f_x = lambda x: (x - 1)**2 + 5
        >>> Optimizer.BFGS(x, f_x, 1000)
        ((0.08925986289978027, 5.000000001826295, array([1.00004274]))

        """

        # start the timer
        start = time.time()
        if np.isscalar(x):
            n = 1
        else:
            n = len(x)
        curr_val = x
        # start with the hessian as the identity matrix
        curr_hessian = np.eye(n)
        fm = forward_mode(x, f_x)
        val, x_der = fm.get_function_value(), fm.get_jacobian()

        # perform BFGS optimization for the number of iterations specified
        for t in range(1, num_iter + 1):
            # compute the new variation to update the current x location
            variation = -alpha * curr_hessian @ x_der
            curr_val = curr_val + variation
            # recalculate the function value and derivative at the updated value
            fm = forward_mode(curr_val, f_x)
            val, x_der1 = fm.get_function_value(), fm.get_jacobian()
            I_mat = np.eye(n)
            y_update, x_update = (x_der1 - x_der).reshape(-1, 1), variation.reshape(-1, 1)
            denom = y_update.T @ x_update
            # compute each matrix individually to determine the amount the hessian will change by
            first_term, sec_term, third_term = (I_mat - x_update @ y_update.T / denom), \
                                               (I_mat - y_update @ x_update.T / denom), \
                                               x_update @ x_update.T / denom

            # update the hessian matrix per iteration (second-order method)
            curr_hessian = first_term @ curr_hessian @ sec_term + third_term
            x_der = x_der1

        # end the timer and compute wall clock time
        end = time.time()
        opt_time = end - start
        return opt_time, val, curr_val



# Authors: Hari Raval, Zongjun Liu, Aditi Memani, Joseph Zuccarelli             #
# Course: AC 207                                                                #
# File: forward_mode.py                                                         #
# Description: Perform forward mode automatic differentiation, enabling a user  #
# to output just the function values, just the derivative values, or both the   #
# function and derviative values in a tuple                                     #
#################################################################################

import numpy as np
from ad_AHJZ.val_derv import val_derv


def combine_vector_inputs(function_list, n_col):
    """
    function to combine a list of values and derivatives for vector cases

    Parameters
    ----------
    function_list: list object of functions
    n_col: integer object that represents the number of columns

    Returns
    -------
    funct_val: an array of values for each input function
    funct_der: an array of derivative values for each function

    Examples
    --------
    # instantiate a seed vector function
    >>> def seed(position):
    >>>     seed_vector = np.zeros(2)
    >>>     seed_vector[position] = 1
    >>>     return seed_vector
    # append all val_derv objects to a simple list
    >>> inputs = [1, 2]
    >>> var_init = []
    >>> for i, value in enumerate(inputs):
    >>>     var_init.append(val_derv(inputs[i], seed(i)))
    # create a function to take the derivative of and combine the inputs to receive a vector output
    >>> func = lambda x,y: np.array([x + y, x * y])
    >>> res = func(*var_init)
    >>> funct_val, funct_der = combine_vector_inputs(res, 2)
    >>> print(funct_val)
    [3. 2.]
    >>> print(funct_der)
    [[1. 1.]
    [2. 1.]]

    """

    # determine the number of functions inputted
    number_functions = len(function_list)

    # arrays to store the function and derivative values
    funct_val = np.empty(number_functions)
    funct_der = np.empty([number_functions, n_col])

    # compute the function and derivative value for each of the input functions
    for i, funct in enumerate(function_list):
        funct_val[i] = funct.val
        funct_der[i] = funct.derv

    return funct_val, funct_der


class forward_mode:
    """
    A class to perform forward mode automatic differentiation, enabling a user
    to output just the function values, just the derivative values, or both the
    function and derviative values in a tuple

    Instance Variables
    ----------
    input_values: a scalar value or a vector of values
    input_function: a scalar function or a vector of functions

    Examples
    --------
    # get function value
    >>> func = lambda x: x**2 + 1
    >>> fm = forward_mode(1, func)
    >>> print(fm.get_function_value())
    2

    # get function derivative
    >>> print(fm.get_jacobian())
    array([2.])

    # get function value and derivative
    >>> print(fm.get_function_value_and_jacobian())
    (2, array([2.]))

    """

    def __init__(self, input_values, input_function):
        self.inputs = input_values
        self.functions = input_function

    def get_function_value(self):
        """
        Parameters
        ----------
        None

        Returns
        -------
        evaluated value of the input function

        Examples
        --------
        # get univariate scalar function value
        >>> func = lambda x: x
        >>> fm = forward_mode(1, func)
        >>> fm.get_function_value()
        1

        # get univariate vector function value
        >>> func = lambda x: (x, 2*x, x**2)
        >>> fm = forward_mode(1, func)
        >>> fm.get_function_value()
        array([1., 2., 1.])

        # get multivariate scalar function value
        >>> func = lambda x, y: x + y
        >>> fm = forward_mode(np.array([1, 1]), func)
        >>> fm.get_function_value()
        2

        # get multivariate vector function value
        >>> func = lambda x, y: (x.log() + y, x + y)
        >>> fm = forward_mode(np.array([1, 1]), func)
        >>> fm.get_function_value()
        array([1., 2.])

        """

        return self.get_function_value_and_jacobian()[0]

    def get_jacobian(self):
        """
        Parameters
        ----------
        None

        Returns
        -------
        jacobian of the input function

        Examples
        --------
        # get univariate scalar function jacobian
        >>> func = lambda x: x
        >>> fm = forward_mode(1, func)
        >>> fm.get_jacobian()
        array([1.])

        # get univariate vector function jacobian
        >>> func = lambda x: (x, 2*x, x**2)
        >>> fm = forward_mode(1, func)
        >>> fm.get_jacobian()
        array([[1.],
               [2.],
               [2.]])

        # get multivariate scalar function jacobian
        >>> func = lambda x, y: x + y
        >>> fm = forward_mode(np.array([1, 1]), func)
        >>> fm.get_jacobian()
        array([1., 1.])

        # get multivariate vector function jacobian
        >>> func = lambda x, y: (x.log() + y, x + y)
        >>> fm = forward_mode(np.array([1, 1]), func)
        >>> fm.get_jacobian()
        array([[1., 1.],
               [1., 1.]])

        """

        return self.get_function_value_and_jacobian()[1]

    def get_function_value_and_jacobian(self):
        """
        Parameters
        ----------
        None

        Returns
        -------
        evaluated value and jacobian of the input function

        Examples
        --------
        # get univariate scalar function value and jacobian
        >>> func = lambda x: x
        >>> fm = forward_mode(1, func)
        >>> fm.get_function_value_and_jacobian()
        (1, array([1.]))

        # get univariate vector function value and jacobian
        >>> func = lambda x: (x, 2*x, x**2)
        >>> fm = forward_mode(1, func)
        >>> fm.get_function_value_and_jacobian()
        (array([1., 2., 1.]), array([[1.],
                                     [2.],
                                     [2.]]))

        # get multivariate scalar function value and jacobian
        >>> func = lambda x, y: x + y
        >>> fm = forward_mode(np.array([1, 1]), func)
        >>> fm.get_function_value_and_jacobian()
        (2, array([1., 1.]))

        # get multivariate vector function value and jacobian
        >>> func = lambda x, y: (x.log() + y, x + y)
        >>> fm = forward_mode(np.array([1, 1]), func)
        >>> fm.get_function_value_and_jacobian()
        (array([1., 2.]), array([[1., 1.],
                                 [1., 1.]]))

        """

        # number of input values = number of variables = number of partial derivatives to compute
        try:
            # find the number of terms to evaluate
            number_res = len(self.inputs)
        # handle the case of having only a single input value: convert the scalar value into a list
        except TypeError:
            number_res = 1
            self.inputs = np.array([self.inputs])

        # initialize the number of partial derivatives
        var_init = [0] * number_res

        # compute the seed vector based on the current position
        def seed(position):
            seed_vector = np.zeros(number_res)
            seed_vector[position] = 1
            return seed_vector

        # initialize each variable into a val_derv() object
        for i, value in enumerate(self.inputs):
            var_init[i] = val_derv(self.inputs[i], seed(i))

        res = self.functions(*var_init)

        # return the results for the scalar case
        try:
            return res.val, res.derv

        # return the results for the vector case
        except AttributeError:
            return combine_vector_inputs(res, number_res)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Nov 11 15:14:15 2021

@author: aditimemani
"""
#from val_derv import val_derv
#from forward_mode import combine_vector_inputs, forward_mode
#from optimizers import Optimizer



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Nov 11 15:03:49 2021

@author: aditimemani
"""

import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()


setuptools.setup(name='ad_AHJZ',
      version='1.7',
      packages= setuptools.find_packages(),
      author="Aditi Memani, Hari Raval, Joseph Zuccarelli, liuzongjun",
      description="A Package that calculates Automatic Differentiation for both Scalar and Vector Inputs",
      #long_description= long_description,
      #long_description_content_type="text/markdown",
      url="https://github.com/cs107-AHJZ/cs107-FinalProject.git",
      classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
      ],
      python_requires='>=3.6',
      install_requires = ['numpy==1.19.3' ]
)



# Authors: Hari Raval, Zongjun Liu, Aditi Memani, Joseph Zuccarelli             #
# Course: AC 207                                                                #
# File: val_derv.py                                                             #
# Description: This class defines a variable object to be used in automatic     #
# differentiation, which encapsulates the cases of both real and dual numbers   #
#################################################################################

import numpy as np


def var_type(x):
    """
    method to check whether input x contains numeric elements or not

    Parameters
    ----------
    x: An object to be checked if the values contained within it are either integers or floats

    Returns
    -------
    True or False depending on whether all of the elements within object x are either integers or floats

    Examples
    --------
    >>> x = 3
    >>> print(var_type(x))
    True
    >>> x = 'a'
    >>> print(var_type(x))
    False

    """

    # if the input object is not a character and is a scalar, it must be numeric
    if not isinstance(x, str) and np.isscalar(x):
        return True
    # if the input object is of an "array" or "list" type, check all contents
    else:
        for i in x:
            # if current element is numeric, move on to the next
            if isinstance(i, (float, int, np.int32, np.int64, np.float64)):
                pass
            else:
                return False

        return True


class val_derv:
    """
    A class representing a custom variable object to be used in automatic differentiation.
    The class encapsulates the cases of both real and dual numbers, containing overloaded dunder methods
    and overloaded elementary functions that are used in forward mode calculation.

    Instance Variables
    ----------
    val: value of the val_derv object
    derv: derivative(s) of the val_derv object

    Returns
    -------
    A val_derv object that contains the value and derivative

    Examples
    --------
    # sample instantiation and use case of val_derv object
    >>> x_1 = val_derv(1, 1)
    >>> x_2 = val_derv(1, 1)
    >>> print(x_1 + x_2)
    Values:2, Derivatives:2

    """

    def __init__(self, val, derv_seed):
        """
        constructor to create a val_derv object represented by a value and derivative

        Parameters
        ----------
        val: An integer or float object that represents the variable value
        derv_seed: An integer or float object that represents the seed value for the variable derivative

        Returns
        -------
        None

        """

        self.val = val
        self.derv = derv_seed

    @property
    def val(self):
        """
        method to retrieve the value attribute of a val_derv object

        Parameters
        ----------
        None

        Returns
        -------
        val attribute of val_derv object

        Examples
        --------
        >>> x = val_derv(1, 1)
        >>> print(x.val)
        1

        """

        return self._val

    @val.setter
    def val(self, val):
        """
        method to set the value attribute of val_derv object

        Parameters
        ----------
        val: A float or integer object that represents the value of the val_derv object

        Returns
        -------
        None

        Raises
        ------
        TypeError
            If input contains non-integer or non-float value
            If input contains a 1D numpy array of non-integer or non-float values

        Examples
        --------
        >>> x = val_derv(1, 1)
        >>> x.val = 2
        2

       """

        if var_type(val):
            self._val = val
        # ensure user does not pass in non-numeric input
        else:
            raise TypeError('ERROR: Input value should be an int or float')

    @property
    def derv(self):
        """
        method to retrieve the derivative attribute of val_derv object

        Parameters
        ----------
        None

        Returns
        -------
        derv attribute of val_derv object

        Examples
        --------
        >>> x = val_derv(1, 1)
        >>> print(x.val)
        1

        """

        return self._derv

    @derv.setter
    def derv(self, derv):
        """
        method to set the derivative attribute of val_derv object

        Parameters
        ----------
        derv: A float/integer object or 1D array of float/integer objects that represents val_derv derivative

        Returns
        -------
        None

        Raises
        ------
        TypeError
            If input contains non-integer or non-float value
            If input contains a 1D numpy array of non-integer or non-float values

        Examples
        --------
        >>> x = val_derv(1, 1)
        >>> x.derv = 2
        2

        """

        # ensure the derivative is numeric before setting val_derv attribute
        if var_type(derv):
            self._derv = derv
        # in the case of a 1D array of derivatives, check each element individually
        elif isinstance(derv, np.ndarray) and len(derv.shape) == 1:
            try:
                derv = derv.astype(float)
            except ValueError:
                raise ValueError('ERROR: Input value should be an int or float')
            self._derv = derv
        # for all other non-numeric cases, raise the appropriate value error
        else:
            raise TypeError('ERROR: Input value must contain an array of ints/floats or be a scalar int/float')

    def __repr__(self):
        """
        method to overload the string representation for val_derv object

        Parameters
        ----------
        None

        Returns
        -------
        Formatted string representation of val_derv object

        Examples
        --------
        >>> print(val_derv(1, 2))
        Values:1, Derivatives:2

        """

        return f'Values:{self.val}, Derivatives:{self.derv}'

    def __add__(self, other):
        """
        method to compute the value and derivative of the addition operation

        Parameters
        ----------
        other: A float/integer object or val_derv object

        Returns
        -------
        A val_derv object that contains the value and derivative of the addition operation

        Examples
        --------
        # add of variables with scalar derivative
        >>> x_1 = val_derv(1, 1)
        >>> x_2 = val_derv(1, 1)
        >>> print(x_1 + x_2)
        Values:2, Derivatives:2

        # add of variables with vector derivative
        >>> x = val_derv(1, np.array([1, 0]))
        >>> y = val_derv(1, np.array([0, 1]))
        >>> print(x + y)
        Values:2, Derivatives:[1 1]

        """

        # perform addition assuming other is a dual number
        try:
            f = self.val + other.val
            f_prime = self.derv + other.derv
        # perform addition when other is a real number
        except AttributeError:
            f = self.val + other
            f_prime = self.derv
        return val_derv(f, f_prime)

    def __sub__(self, other):
        """
        method to compute the value and derivative of the subtraction operation

        Parameters
        ----------
        other: A float/integer object or val_derv object

        Returns
        -------
        A val_derv object that contains the value and derivative of the subtraction operation

        Examples
        --------
        # subtraction of variable with scalar derivative
        >>> x_1 = val_derv(1, 1)
        >>> x_2 = val_derv(1, 1)
        >>> print(x_1 - x_2)
        Values:0, Derivatives:0

        # subtraction of variable with vector derivative
        >>> x = val_derv(1, np.array([1, 0]))
        >>> y = val_derv(1, np.array([0, 1]))
        >>> print(x - y)
        Values:0, Derivatives:[ 1 -1]

        """

        # perform subtraction assuming other is a dual number
        try:
            f = self.val - other.val
            f_prime = self.derv - other.derv
        # perform subtraction when other is a real number
        except AttributeError:
            f = self.val - other
            f_prime = self.derv
        return val_derv(f, f_prime)

    def __mul__(self, other):
        """
        method to compute the value and derivative of the multiplication operation

        Parameters
        ----------
        other: A float/integer object or val_derv object

        Returns
        -------
        A val_derv object that contains the value and derivative of the multiplication operation

        Examples
        --------
        # multiplication of variable with scalar derivative
        >>> x_1 = val_derv(1, 1)
        >>> x_2 = val_derv(1, 1)
        >>> print(x_1 * x_2)
        Values:1, Derivatives:2

        # multiplication of variable with vector derivative
        >>> x = val_derv(1, np.array([1, 0]))
        >>> y = val_derv(1, np.array([0, 1]))
        >>> print(x * y)
        Values:1, Derivatives:[1 1]

        """

        # perform multiplication assuming other is a dual number
        try:
            f = self.val * other.val
            f_prime = self.val * other.derv + self.derv * other.val
        # perform multiplication when other is a real number
        except AttributeError:
            f = self.val * other
            f_prime = self.derv * other
        return val_derv(f, f_prime)

    def __truediv__(self, other):
        """
        method to compute the value and derivative of the division operation

        Parameters
        ----------
        other: A float/integer object or val_derv object

        Returns
        -------
        A val_derv object that contains the value and derivative of the division operation

        Raises
        ------
        ZeroDivisionError if denominator in division is zero

        Examples
        --------
        # true division of variable with scalar derivative
        >>> x_1 = val_derv(1, 1)
        >>> x_2 = val_derv(1, 1)
        >>> print(x_1 / x_2)
        Values:1.0, Derivatives:0.0

        # true division of variable with vector derivative
        >>> x = val_derv(1, np.array([1, 0]))
        >>> y = val_derv(1, np.array([0, 1]))
        >>> print(x / y)
        Values:1.0, Derivatives:[ 1. -1.]

        # ZeroDivisionError if denominator in division is zero
        >>> x_1 = val_derv(1, 1)
        >>> x_2 = val_derv(0, 1)
        >>> print(x_1 / x_2)
        ZeroDivisionError: ERROR: Denominator in division should not be 0

        # ZeroDivisionError if denominator in division is zero
        >>> x_1 = val_derv(1, 1)
        >>> x_2 = 0
        >>> print(x_1 / x_2)
        ZeroDivisionError: ERROR: Denominator in division should not be 0

        """

        # perform division assuming other is a dual number
        try:
            # ensure the user does not divide by 0
            if other.val == 0:
                raise ZeroDivisionError("ERROR: Denominator in division should not be 0")
            f = self.val / other.val
            f_prime = (self.derv * other.val - self.val * other.derv) / other.val ** 2
            return val_derv(f, f_prime)
        # perform division when other is a real number
        except AttributeError:
            # ensure the user does not divide by 0
            if other == 0:
                raise ZeroDivisionError("ERROR: Denominator in division should not be 0")
            f = self.val / other
            f_prime = self.derv / other
            return val_derv(f, f_prime)

    def __neg__(self):
        """
        method to compute the value and derivative of the negation operation

        Parameters
        ----------
        None

        Returns
        -------
        A val_derv object that contains the value and derivative of the negation operation

        Examples
        --------
        # negation of variable with scalar derivative
        >>> x = val_derv(1, 1)
        >>> print(-x)
        Values:-1, Derivatives:-1

        # negation of variable with vector derivative
        >>> x = val_derv(1, np.array([1, 0]))
        >>> print(-x)
        Values:-1, Derivatives:[-1  0]

        """

        return val_derv(-1 * self.val, -1 * self.derv)

    def __pow__(self, other):
        """
        method to compute the value and derivative of the power operation

        Parameters
        ----------
        other: A float/integer object or val_derv object

        Returns
        -------
        A val_derv object that contains the value and derivative of the power operation

        Raises
        ------
        ValueError
            If negative number is raised to a fraction with an even denominator
            If exponent is less than 1 and differentiation occurs at 0

        Examples
        --------
        # power of variable with scalar derivative
        >>> x_1 = val_derv(1, 1)
        >>> x_2 = val_derv(1, 1)
        >>> print(x_1 ** x_2)
        Values:1, Derivatives:1.0

        # power of variable with vector derivative
        >>> x = val_derv(1, np.array([1, 0]))
        >>> y = val_derv(1, np.array([0, 1]))
        >>> print(x ** y)
        Values:1, Derivatives:[1. 0.]

        # ValueError if negative number is raised to a fraction with an even denominator
        >>> x_1 = val_derv(-1, 1)
        >>> x_2 = val_derv(0.5, 1)
        >>> print(x_1 ** x_2)
        ValueError: ERROR: Attempted to raise a negative number to a fraction with even denominator

        # ValueError if exponent is less than 1 and differentiation occurs at 0
        >>> x_1 = val_derv(0, 1)
        >>> x_2 = val_derv(0.5, 1)
        >>> print(x_1 ** x_2)
        ValueError: ERROR: Attempted to find derivative at 0 when exponent is less than 1

        """

        # perform power operation assuming other is a dual number
        try:
            # ensure user does not raise a negative number to a fraction with an even denominator
            if self.val < 0 and other.val % 1 != 0 and other.val.as_integer_ratio()[1] % 2 == 0:
                raise ValueError("ERROR: Attempted to raise a negative number to a fraction with even denominator")
            # ensure user does not have a 0 derivative when the exponent is less than 1
            if self.val == 0 and other.val < 1:
                raise ValueError("ERROR: Attempted to find derivative at 0 when exponent is less than 1")

            f = self.val ** other.val
            # compute the derivative power rule for a dual number exponent
            f_prime = (self.val ** (other.val - 1)) * self.derv * other.val + (
                    self.val ** other.val) * other.derv * np.log(self.val)
            return val_derv(f, f_prime)

        # perform power operation when other is a real number
        except AttributeError:
            # ensure user does not raise a negative number to a fraction with an even denominator
            if self.val < 0 and other % 1 != 0 and other.as_integer_ratio()[1] % 2 == 0:
                raise ValueError("ERROR: Attempted to raise a negative number to a fraction with even denominator")
            # ensure user does not have a 0 derivative when the exponent is less than 1
            if self.val == 0 and other < 1:
                raise ValueError("ERROR: Attempted to find derivative at 0 when exponent is less than 1")

            f = self.val ** other
            # compute the derivative power rule for a real number exponent
            f_prime = other * self.val ** (other - 1)
            return val_derv(f, self.derv * f_prime)

    def __radd__(self, other):
        """
        method to compute the value and derivative of the reverse addition operation

        Parameters
        ----------
        other: A float/integer object

        Returns
        -------
        A val_derv object that contains the value and derivative of the reverse addition operation

        Examples
        --------
        # reverse addition of variable with scalar derivative
        >>> x_1 = 1
        >>> x_2 = val_derv(1, 1)
        >>> print(x_1 + x_2)
        Values:2, Derivatives:1

        # reverse addition of variable with vector derivative
        >>> x = 1
        >>> y = val_derv(1, np.array([0, 1]))
        >>> print(x + y)
        Values:2, Derivatives:[0 1]

        """

        return self + other

    def __rsub__(self, other):
        """
        method to compute the value and derivative of the reverse subtraction operation

        Parameters
        ----------
        other: A float/integer object

        Returns
        -------
        A val_derv object that contains the value and derivative of the reverse subtraction operation

        Examples
        --------
        # reverse subtraction of variable with scalar derivative
        >>> x_1 = 1
        >>> x_2 = val_derv(1, 1)
        >>> print(x_1 - x_2)
        Values:0, Derivatives:-1

        # reverse subtraction of variable with vector derivative
        >>> x = 1
        >>> y = val_derv(1, np.array([0, 1]))
        >>> print(x - y)
        Values:0, Derivatives:[ 0 -1]

        """

        return other + (-self)

    def __rmul__(self, other):
        """
        method to compute the value and derivative of the reverse multiplication operation

        Parameters
        ----------
        other: A float/integer object

        Returns
        -------
        A val_derv object that contains the value and derivative of the reverse multiplication operation

        Examples
        --------
        # reverse multiplication of variable with scalar derivative
        >>> x_1 = 1
        >>> x_2 = val_derv(1, 1)
        >>> print(x_1 * x_2)
        Values:1, Derivatives:1

        # reverse multiplication of variable with vector derivative
        >>> x = 1
        >>> y = val_derv(1, np.array([0, 1]))
        >>> print(x * y)
        Values:1, Derivatives:[0 1]

        """

        return self * other

    def __rtruediv__(self, other):
        """
        method to compute the value and derivative of the reverse division operation

        Parameters
        ----------
        other: A float/integer object

        Returns
        -------
        A val_derv object that contains the value and derivative of the reverse division operation

        Raises
        ------
        ZeroDivisionError if denominator in division is zero

        Examples
        --------
        # reverse true division of variable with scalar derivative
        >>> x_1 = 1
        >>> x_2 = val_derv(1, 1)
        >>> print(x_1 / x_2)
        Values:1.0, Derivatives:-1.0

        # reverse true division of variable with vector derivative
        >>> x = 1
        >>> y = val_derv(1, np.array([0, 1]))
        >>> print(x / y)
        Values:1.0, Derivatives:[ 0. -1.]

        # ZeroDivisionError if denominator in division is zero
        >>> x_1 = 1
        >>> x_2 = val_derv(0, 1)
        >>> print(x_1 / x_2)
        ZeroDivisionError: ERROR: Denominator in division should not be 0

        """

        # ensure the user does not attempt to divide by 0
        if self.val == 0:
            raise ZeroDivisionError("ERROR: Denominator in division should not be 0")

        # compute the value and derivative of the reverse division analogus to forward direction
        f = other / self.val
        f_prime = - other * self.derv / self.val ** 2
        return val_derv(f, f_prime)

    def __rpow__(self, other):
        """
        method to compute the value and derivative of the reverse power operation

        Parameters
        ----------
        other: A float/integer object

        Returns
        -------
        A val_derv object that contains the value and derivative of the reverse power operation

        Examples
        --------
        # reverse power of variable with scalar derivative
        >>> x_1 = 1
        >>> x_2 = val_derv(1, 1)
        >>> print(x_1 ** x_2)
        Values:1, Derivatives:0.0

        # reverse power of variable with vector derivative
        >>> x = 1
        >>> y = val_derv(1, np.array([0, 1]))
        >>> print(x ** y)
        Values:1, Derivatives:[0. 0.]

        """

        # compute value and derivative of the reverse power function analogous to forward direction
        f = other ** self.val
        f_prime = (other ** self.val) * self.derv * np.log(other)
        return val_derv(f, f_prime)

    def __eq__(self, other):
        """
        method to check whether the value and derivative of val_derv objects are equal

        Parameters
        ----------
        other: A val_derv object

        Returns
        -------
        A tuple of boolean entries where the first entry determines if the function values are equal and the second
        entry determines if the derivative values are equal

        Examples
        --------
        # equality of val_derv objects with the same contents
        >>> val_derv1 = val_derv(1, 10)
        >>> val_derv2 = val_derv(1, 10)
        >>> print(val_derv1 == val_derv2)
        (True, True)

        # equality of val_derv objects with partially different contents
        >>> val_derv1 = val_derv(1, 10)
        >>> val_derv2 = val_derv(2, 10)
        >>> print(val_derv1 == val_derv2)
        (False, True)

        # equality of val_derv objects with all different contents
        >>> val_derv1 = val_derv(2, 1)
        >>> val_derv2 = val_derv(1, 2)
        >>> print(val_derv1 == val_derv2)
        (False, False)

        """

        # check if val_derv values are equal
        try:
            value_eq = all(self.val == other.val)
        except TypeError:
            value_eq = True if self.val == other.val else False

        # check if val_derv derivatives are equal
        try:
            derivative_eq = all(self.derv == other.derv)
        except TypeError:
            derivative_eq = True if self.derv == other.derv else False

        return value_eq, derivative_eq

    def __ne__(self, other):
        """
        method to check whether the value and derivative of val_derv objects are not equal

        Parameters
        ----------
        other: A val_derv object

        Returns
        -------
        A tuple of boolean entries where the first entry determines if the function values are not equal and
        the second entry determines if the derivative values are not equal

        Examples
        --------
        # non-equality of val_derv objects with the same contents
        >>> val_derv1 = val_derv(1, 10)
        >>> val_derv2 = val_derv(1, 10)
        >>> print(val_derv1 != val_derv2)
        (False, False)

        # non-equality of val_derv objects with partially different contents
        >>> val_derv1 = val_derv(1, 10)
        >>> val_derv2 = val_derv(2, 10)
        >>> print(val_derv1 != val_derv2)
        (True, False)

        # non-equality of val_derv objects with all different contents
        >>> val_derv1 = val_derv(2, 1)
        >>> val_derv2 = val_derv(1, 2)
        >>> print(val_derv1 != val_derv2)
        (True, True)

        """

        # check if val_derv values are not equal
        try:
            value_eq = all(self.val != other.val)
        except TypeError:
            value_eq = True if self.val != other.val else False

        # check if val_derv derivatives are not equal
        try:
            derivative_eq = all(self.derv != other.derv)
        except TypeError:
            derivative_eq = True if self.derv != other.derv else False

        return value_eq, derivative_eq

    def sqrt(self):
        """
        method to compute the value and derivative of the square root function

        Parameters
        ----------
        None

        Returns
        -------
        A val_derv object that contains the value and derivative of the square root function

        Examples
        --------
        # square root of variable with scalar derivative
        >>> x = val_derv(1, 1)
        >>> print(x.sqrt())
        Values:1.0, Derivatives:0.5

        # square root of variable with vector derivative
        >>> x = val_derv(1, np.array([1, 0]))
        >>> print(x.sqrt())
        Values:1.0, Derivatives:[0.5 0. ]

        """

        return self.__pow__(0.5)

    def log(self, base=None):
        """
        method to compute the value and derivative of logarthmic function
        Parameters

        Parameters
        ----------
        base: A float/integer object that represents the base of the logarithm (default logarithmic base is None)

        Returns
        -------
        A val_derv object that contains the value and derivative of the logarithmic function

        Raises
        ------
        ValueError
            If self.val is less than or equal to zero
            If input base is less than or equal to zero
            If input base is equal to one

        Examples
        --------
        # ValueError if self.val is less than or equal to zero
        >>> x = val_derv(0, 1)
        >>> print(x.log())
        ValueError: ERROR: Value for log should be greater than 0
        >>> x = val_derv(-1, 1)
        >>> print(x.log())
        ValueError: ERROR: Value for log should be greater than 0

        # ValueError if input base is less than or equal to zero or equal to 1
        >>> x = val_derv(1, 1)
        >>> print(x.log(base = -1))
        ValueError: ERROR: LOG base should be greater than 0 and not equal to 1
        >>> x = val_derv(1, 1)
        >>> print(x.log(base = 1))
        ValueError: ERROR: LOG base should be greater than 0 and not equal to 1
        >>> x = val_derv(1, 1)
        >>> print(x.log(base = 0))
        ValueError: ERROR: LOG base should be greater than 0 and not equal to 1

        # logarithm of variable with scalar derivative
        >>> x = val_derv(1, 1)
        >>> print(x.log())
        Values:0.0, Derivatives:1.0

        # logarithm of variable with vector derivative
        >>> x = val_derv(1, np.array([1, 0]))
        >>> print(x.log())
        Values:0.0, Derivatives:[1. 0.]

        """

        # ensure user does not pass a negative value to take the log of
        if self.val <= 0:
            raise ValueError("ERROR: Value for log should be greater than 0")

        # if the default base is used, proceed without checking base conditions
        if base is None:
            f = np.log(self.val)
            f_prime = 1 / self.val
            return val_derv(f, self.derv * f_prime)
        # ensure the user specifies a valid base before computing the log value and derivative
        else:
            if base <= 0 or base == 1:
                raise ValueError("ERROR: LOG base should be greater than 0 and not equal to 1")
            f = np.log(self.val) / np.log(base)
            f_prime_deno = self.val * np.log(base)
            f_prime = 1 / f_prime_deno
            return val_derv(f, self.derv * f_prime)

    def exp(self):
        """
        method to compute the value and derivative of exponential function

        Parameters
        ----------
        None

        Returns
        -------
        A val_derv object that contains the value and derivative of the exponential function

        Examples
        --------
        # exponential of variable with scalar derivative
        >>> x = val_derv(0, 1)
        >>> print(x.exp())
        Values:1.0, Derivatives:1.0

        # exponential of variable with vector derivative
        >>> x = val_derv(0, np.array([1, 0]))
        >>> print(x.exp())
        Values:1.0, Derivatives:[1. 0.]

        """

        # compute the value and derivative of the exponential function for any input
        f = np.exp(self.val)
        f_prime = np.exp(self.val)
        return val_derv(f, self.derv * f_prime)

    def sin(self):
        """
        method to compute the value and derivative of the sine function

        Parameters
        ----------
        None

        Returns
        -------
        A val_derv object that contains the value and derivative of the sine function

        Examples
        --------
        # sine of variable with scalar derivative
        >>> x = val_derv(0, 1)
        >>> print(x.sin())
        Values:0.0, Derivatives:1.0

        # sine of variable with vector derivative
        >>> x = val_derv(0, np.array([1, 0]))
        >>> print(x.sin())
        Values:0.0, Derivatives:[1. 0.]

        """

        # compute the value and derivative of the sine function for any input
        f = np.sin(self.val)
        f_prime = np.cos(self.val)
        return val_derv(f, self.derv * f_prime)

    def cos(self):
        """
        method to compute the value and derivative of the cosine function

        Parameters
        ----------
        None

        Returns
        -------
        A val_derv object that contains the value and derivative of the cosine function

        Examples
        --------
        # cosine of variable with scalar derivative
        >>> x = val_derv(0, 1)
        >>> print(x.cos())
        Values:1.0, Derivatives:-0.0

        # cosine of variable with vector derivative
        >>> x = val_derv(0, np.array([1, 0]))
        >>> print(x.cos())
        Values:1.0, Derivatives:[-0. -0.]

        """

        # compute the value and derivative of the cosine function for any input
        f = np.cos(self.val)
        f_prime = - np.sin(self.val)
        return val_derv(f, self.derv * f_prime)

    def tan(self):
        """
        method to compute the value and derivative of the tangent function

        Parameters
        ----------
        None

        Returns
        -------
        A val_derv object that contains the value and derivative of the tangent function

        Raises
        ------
        ValueError if input is an odd multiple of pi/2

        Examples
        --------
        # tangent of variable with scalar derivative
        >>> x = val_derv(0, 1)
        >>> print(x.tan())
        Values:0.0, Derivatives:1.0

        # tangent of variable with vector derivative
        >>> x = val_derv(0, np.array([1, 0]))
        >>> print(x.tan())
        Values:0.0, Derivatives:[1. 0.]

        # ValueError if input is an odd multiple of pi/2
        >>> x = val_derv(np.pi / 2, 1)
        >>> print(x.tan())
        ValueError: ERROR: Input to tan should not be an odd mutiple of pi/2

        """

        # ensure the user does not input an odd multiple of pi divided by 2
        if (self.val / (np.pi / 2)) % 2 == 1:
            raise ValueError("ERROR: Input to tan should not be an odd mutiple of pi/2")

        # compute the value and derivative of the tangent function for a valid input
        f = np.tan(self.val)
        f_prime = 1 / np.cos(self.val) ** 2
        return val_derv(f, self.derv * f_prime)

    def sinh(self):
        """
        method to compute the value and derivative of the hyperbolic sine function

        Parameters
        ----------
        None

        Returns
        -------
        A val_derv object that contains the value and derivative of the hyperbolic sine function

        Examples
        --------
        # hyperbolic sine of variable with scalar derivative
        >>> x = val_derv(0, 1)
        >>> print(x.sinh())
        Values:0.0, Derivatives:1.0

        # hyperbolic sine of variable with vector derivative
        >>> x = val_derv(0, np.array([1, 0]))
        >>> print(x.sinh())
        Values:0.0, Derivatives:[1. 0.]

        """

        # compute the value and derivative of the hyperbolic sine for any input
        f = np.sinh(self.val)
        f_prime = np.cosh(self.val)
        return val_derv(f, self.derv * f_prime)

    def cosh(self):
        """
        method to compute the value and derivative of the hyperbolic cosine function

        Parameters
        ----------
        None

        Returns
        -------
        A val_derv object that contains the value and derivative of the hyperbolic cosine function

        Examples
        --------
        # hyperbolic cosine of variable with scalar derivative
        >>> x = val_derv(0, 1)
        >>> print(x.cosh())
        Values:1.0, Derivatives:0.0

        # hyperbolic cosine of variable with vector derivative
        >>> x = val_derv(0, np.array([1, 0]))
        >>> print(x.cosh())
        Values:1.0, Derivatives:[0. 0.]

        """

        # compute the value and derivative of the hyperbolic cosine function for any input
        f = np.cosh(self.val)
        f_prime = np.sinh(self.val)
        return val_derv(f, self.derv * f_prime)

    def tanh(self):
        """
        method to compute the value and derivative of the hyperbolic tangent function

        Parameters
        ----------
        None

        Returns
        -------
        A val_derv object that contains the value and derivative of the hyperbolic tangent function

        Examples
        --------
        # hyperbolic tangent of variable with scalar derivative
        >>> x = val_derv(0, 1)
        >>> print(x.tanh())
        Values:0.0, Derivatives:1.0

        # hyperbolic tangent of variable with vector derivative
        >>> x = val_derv(0, np.array([1, 0]))
        >>> print(x.tanh())
        Values:0.0, Derivatives:[1. 0.]

        """

        # compute the value and derivative of the hyperbolic tangent function for any input
        f = np.tanh(self.val)
        f_prime = 1 / (np.cosh(self.val) ** 2)
        return val_derv(f, self.derv * f_prime)

    def arcsin(self):
        """
        method to compute the value and derivative of the inverse sine function

        Parameters
        ----------
        None

        Returns
        -------
        A val_derv object that contains the value and derivative of the inverse sine function

        Raises
        ------
        ValueError if input is not contained within the interval [-1,1]

        Examples
        --------
        # inverse sine of variable with scalar derivative
        >>> x = val_derv(0, 1)
        >>> print(x.arcsin())
        Values:0.0, Derivatives:1.0

        # inverse sine of variable with vector derivative
        >>> x = val_derv(0, np.array([1, 0]))
        >>> print(x.arcsin())
        Values:0.0, Derivatives:[1. 0.]

        # ValueError for input outside of the interval -1 to 1
        >>> x = val_derv(2, 1)
        >>> print(x.arcsin())
        ValueError: ERROR: Input to arcsin() should be between -1 and 1

        """

        # ensure the user passes in an input between -1 and 1
        if -1 >= self.val or self.val >= 1:
            raise ValueError("ERROR: Input to arcsin() should be between -1 and 1")
        # compute the value and derivative of the inverse sine function for a valid input
        f = np.arcsin(self.val)
        f_prime = 1 / (1 - self.val ** 2) ** 0.5
        return val_derv(f, self.derv * f_prime)

    def arccos(self):
        """
        method to compute the value and derivative of the inverse cosine function

        Parameters
        ----------
        None

        Returns
        -------
        A val_derv object that contains the value and derivative of the inverse cosine function

        Raises
        ------
        ValueError will be raised if input is not contained within the interval [-1,1]

        Examples
        --------
        # inverse cosine of variable with scalar derivative
        >>> x = val_derv(0, 1)
        >>> print(x.arccos())
        Values:1.5707963267948966, Derivatives:-1.0

        # inverse cosine of variable with vector derivative
        >>> x = val_derv(0, np.array([1, 0]))
        >>> print(x.arccos())
        Values:1.5707963267948966, Derivatives:[-1. -0.]

        # ValueError for input outside of the interval -1 to 1
        >>> x = val_derv(2, 1)
        >>> print(x.arccos())
        ValueError: ERROR: Input to arccos() should be between -1 and 1

        """

        # ensure the user passes in an input between -1 and 1
        if -1 >= self.val or self.val >= 1:
            raise ValueError("ERROR: Input to arccos() should be between -1 and 1")
        # compute the value and derivative of the inverse cosine function for a valid input
        f = np.arccos(self.val)
        f_prime = - 1 / (1 - self.val ** 2) ** 0.5
        return val_derv(f, self.derv * f_prime)

    def arctan(self):
        """
        method to compute the value and derivative of the inverse tangent function

        Parameters
        ----------
        None

        Returns
        -------
        A val_derv object that contains the value and derivative of the inverse tangent function

        Examples
        --------
        # inverse tangent of variable with scalar derivative
        >>> x = val_derv(1, 1)
        >>> print(x.arctan())
        Values:0.7853981633974483, Derivatives:0.5

        # inverse tangent of variable with vector derivative
        >>> x = val_derv(1, np.array([1, 0]))
        >>> print(x.arctan())
        Values:0.7853981633974483, Derivatives:[0.5 0. ]

        """

        # compute the value and derivative of the inverse tangent function for a valid input
        f = np.arctan(self.val)
        f_prime = 1 / (1 + self.val ** 2)
        return val_derv(f, self.derv * f_prime)

    def logistic(self):
        """
        method to compute the value and derivative of the logistic function

        Parameters
        ----------
        None

        Returns
        -------
        A val_derv object that contains the value and derivative of the logistic function

        Examples
        --------
        # logistic of variable with scalar derivative
        >>> x = val_derv(1, 1)
        >>> print(x.logistic())
        Values:0.7310585786300049, Derivatives:0.19661193324148188

        # logistic of variable with vector derivative
        >>> x = val_derv(1, np.array([1, 0]))
        >>> print(x.logistic())
        Values:0.7310585786300049, Derivatives:[0.19661193 0. ]

        """

        return 1 / ((-self).exp() + 1)

